using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using CodeHollow.FeedReader;
using Serilog;
using SharpRss.Models;
using ToolQit.Extensions;

namespace SharpRss.Services
{
    /// <summary>
    /// Managing RSS feeds and groups.
    /// </summary>
    public class RssService : IDisposable
    {
        public RssService()
        {
            SetupTestCategoriesAndFeedsAsync();   
        }
        private readonly DatabaseService _dbService = new DatabaseService();


		public async Task<bool> CreateGroupAsync(GroupModel group) => await _dbService.SetGroupAsync(group);
		public async Task<HashSet<GroupModel>> GetGroupsAsync() => await _dbService.GetGroupsAsync();

		public async Task<bool> AddFeed(string rssUrl, GroupModel? group = null)
        {
            bool result = false;
            Feed fetched = await FetchFeed(rssUrl);
            if (fetched == null)
                return result;
			FeedModel feedModel = new FeedModel(rssUrl)
            {
                GroupId = group?.Id,
                FeedType = fetched.Type.ToString(),
                Description = fetched.Description,
                Language = fetched.Language,
                Copyright = fetched.Copyright,
                DateAdded = DateTimeOffset.Now,
                LastUpdated = DateTimeOffset.Now,
                ImageUrl = fetched.ImageUrl,
            };
            result = await _dbService.SetFeedAsync(feedModel);
            if (!result)
                return result;
            if (await AddFeedItems(fetched.Items, feedModel) == 0)
                Log.Warning($"No feed items added to feed: {feedModel.Url}");
			return result;
        }
        public async Task<HashSet<FeedModel>> GetFeedsAsync(GroupModel? group = null) => await _dbService.GetFeedsAsync(group);
        public async Task<HashSet<FeedModel>> GetUngroupedFeedsAsync() => await _dbService.GetFeedsAsync(new GroupModel() { Id = "" });
		private async Task<int> AddFeedItems(IList<FeedItem> items, FeedModel feedModel)
		{
            int result = 0;
            if (!items.Any())
                return result;
            HashSet<FeedItemModel> itemModels = new HashSet<FeedItemModel>();
            foreach (FeedItem item in items)
            {
                itemModels.Add(new FeedItemModel()
                {
                    Id = item.Id,
                    FeedId = feedModel.Id,
                    Title = item.Title,
                    Description = item.Description,
                    Link = item.Link,
                    LastUpdated = DateTimeOffset.Now,
                    PublishingDate = item.PublishingDate != null ? new DateTimeOffset((DateTime)item.PublishingDate) : null,
                    Author = item.Author,
                    Categories = item.Categories.ToArray(),
                    Content = item.Content
                });
            }
            result = await _dbService.SetFeedItemsAsync(itemModels);
			return result;
		}
		private async Task<Feed> FetchFeed(string url)
        {
            var urls = await FeedReader.ParseFeedUrlsAsStringAsync(url);
            string feedurl = url;
            if (urls.Any())
                feedurl = urls.First();
            return await FeedReader.ReadAsync(feedurl);
        }
        private async void SetupTestCategoriesAndFeedsAsync()
        {
            var groups = await GetGroupsAsync();
            GroupModel testGroup = groups.Single(x => x.Name == "Test");
            var res = await AddFeed("http://fedoramagazine.org/feed/", testGroup);
            res = await AddFeed("https://www.nasa.gov/rss/dyn/breaking_news.rss", testGroup);
            res = await AddFeed("https://journals.plos.org/plosone/feed/atom", testGroup);
            res = await AddFeed("https://itsfoss.com/feed", testGroup);

            /*bool result = await _dbService.SetGroupAsync(new GroupModel() { Name = "News" });
            result = await _dbService.SetGroupAsync(new GroupModel() { Name = "Tech" });
            result = await _dbService.SetGroupAsync(new GroupModel() { Name = "Science" });
            result = await _dbService.SetGroupAsync(new GroupModel() { Name = "Test" });*/
            
            /*GroupModel? editGroup = await _dbService.GetGroupsAsync("Test");
            if (editGroup != null)
            {
                bool result2 = await _dbService.RemoveGroupAsync(editGroup);
            }*/
            /*await _dbService.AddCategoriesAsync(new HashSet<CategoryModel>()
            {
                new CategoryModel() { Name = "All" },
                new CategoryModel() { Name = "RSS" },
                new CategoryModel() { Name = "Tech" },
                new CategoryModel() { Name = "News" }
            });
            await _dbService.AddFeedsAsync(new HashSet<FeedModel>()
            {
                new FeedModel("http://fedoramagazine.org/feed/"),
                new FeedModel("https://www.nasa.gov/rss/dyn/breaking_news.rss"),
                new FeedModel("https://journals.plos.org/plosone/feed/atom"),
                new FeedModel("https://itsfoss.com/feed")
            });*/
        }

		public void Dispose()
		{
			_dbService.Dispose();
		}
	}
}